@@ -72,20 +72,20 @@ void TestXdgSession::initTestCase()
 
     kwinApp()->start();
     Test::setOutputConfig({
-        QRect(0, 0, 1280, 1024),
-        QRect(1280, 0, 1280, 1024),
+        Rect(0, 0, 1280, 1024),
+        Rect(1280, 0, 1280, 1024),
     });
     const auto outputs = workspace()->outputs();
     QCOMPARE(outputs.count(), 2);
-    QCOMPARE(outputs[0]->geometry(), QRect(0, 0, 1280, 1024));
-    QCOMPARE(outputs[1]->geometry(), QRect(1280, 0, 1280, 1024));
+    QCOMPARE(outputs[0]->geometry(), Rect(0, 0, 1280, 1024));
+    QCOMPARE(outputs[1]->geometry(), Rect(1280, 0, 1280, 1024));
 }
 
 void TestXdgSession::init()
 {
     Test::setOutputConfig({
-        QRect(0, 0, 1280, 1024),
-        QRect(1280, 0, 1280, 1024),
+        Rect(0, 0, 1280, 1024),
+        Rect(1280, 0, 1280, 1024),
     });
 
     QVERIFY(Test::setupWaylandConnection(Test::AdditionalWaylandInterface::XdgSessionV1));
@@ -364,22 +364,22 @@ void TestXdgSession::restoreOffscreenPosition()
 
     restoreTemplate({
         .setup = [outputs](Window *window) {
-            QCOMPARE(outputs[1]->geometry(), QRect(1280, 0, 1280, 1024));
-            window->move(QPointF(1280 + 42, 42));
-            QCOMPARE(window->output(), outputs[1]);
-        },
+        QCOMPARE(outputs[1]->geometry(), Rect(1280, 0, 1280, 1024));
+        window->move(QPointF(1280 + 42, 42));
+        QCOMPARE(window->output(), outputs[1]);
+    },
         .between = [outputs]() {
-            OutputConfiguration configuration;
-            {
-                auto changeSet = configuration.changeSet(outputs[1]->backendOutput());
-                changeSet->enabled = false;
-            }
-            workspace()->applyOutputConfiguration(configuration);
-        },
+        OutputConfiguration configuration;
+        {
+            auto changeSet = configuration.changeSet(outputs[1]->backendOutput());
+            changeSet->enabled = false;
+        }
+        workspace()->applyOutputConfiguration(configuration);
+    },
         .restore = [outputs](Window *window) {
-            QCOMPARE(window->pos(), QPointF(0, 0));
-            QCOMPARE(window->output(), outputs[0]);
-        },
+        QCOMPARE(window->pos(), QPointF(0, 0));
+        QCOMPARE(window->output(), outputs[0]);
+    },
     });
 }
 
