@@ -129,7 +129,7 @@ bool DrmOutput::presentAsync(OutputLayer *layer, std::optional<std::chrono::nano
     return m_pipeline->presentAsync(layer, allowedVrrDelay);
 }
 
-QList<std::shared_ptr<OutputMode>> DrmOutput::getModes() const
+QList<std::shared_ptr<OutputMode>> DrmOutput::getModes(const State &state) const
 {
     const auto drmModes = m_pipeline->connector()->modes();
 
@@ -138,6 +138,9 @@ QList<std::shared_ptr<OutputMode>> DrmOutput::getModes() const
     for (const auto &drmMode : drmModes) {
         ret.append(drmMode);
     }
+    for (const auto &custom : state.customModes) {
+        ret.append(m_pipeline->connector()->generateMode(custom.size, custom.refreshRate / 1000.0f, custom.flags | OutputMode::Flag::Custom));
+    }
     return ret;
 }
 
@@ -172,7 +175,7 @@ void DrmOutput::updateConnectorProperties()
     updateInformation();
 
     State next = m_state;
-    next.modes = getModes();
+    next.modes = getModes(next);
     if (!next.currentMode) {
         // some mode needs to be set
         next.currentMode = next.modes.constFirst();
@@ -486,6 +489,7 @@ void DrmOutput::applyQueuedChanges(const std::shared_ptr<OutputChangeSet> &props
     next.enabled = props->enabled.value_or(m_state.enabled) && m_pipeline->crtc();
     next.position = props->pos.value_or(m_state.position);
     next.scale = props->scale.value_or(m_state.scale);
+    next.scaleSetting = props->scaleSetting.value_or(m_state.scaleSetting);
     next.transform = props->transform.value_or(m_state.transform);
     next.manualTransform = props->manualTransform.value_or(m_state.manualTransform);
     next.currentMode = m_pipeline->mode();
@@ -523,10 +527,15 @@ void DrmOutput::applyQueuedChanges(const std::shared_ptr<OutputChangeSet> &props
     next.automaticMaxBitsPerColorLimit = decideAutomaticBpcLimit();
     next.edrPolicy = props->edrPolicy.value_or(m_state.edrPolicy);
     next.dpmsMode = props->dpmsMode.value_or(m_state.dpmsMode);
+    if (props->customModes.has_value()) {
+        next.customModes = *props->customModes;
+        next.modes = getModes(next);
+    }
     next.originalColorDescription = createColorDescription(next);
     next.colorDescription = applyNightLight(next.originalColorDescription, m_sRgbChannelFactors);
     next.sharpnessSetting = props->sharpness.value_or(m_state.sharpnessSetting);
     next.priority = props->priority.value_or(m_state.priority);
+    next.deviceOffset = props->deviceOffset.value_or(m_state.deviceOffset);
     tryKmsColorOffloading(next);
     maybeScheduleRepaints(next);
     if (next.dpmsMode != m_state.dpmsMode) {
