@@ -27,6 +27,7 @@
 #include "opengl/glplatform.h"
 #include "qpainter/qpainterbackend.h"
 #include "renderloopdrivenqanimationdriver.h"
+#include "scene/cursoritem.h"
 #include "scene/itemrenderer_opengl.h"
 #include "scene/itemrenderer_qpainter.h"
 #include "scene/surfaceitem.h"
@@ -106,15 +107,6 @@ BackendOutput *Compositor::findOutput(RenderLoop *loop) const
     return nullptr;
 }
 
-LogicalOutput *Compositor::findLogicalOutput(BackendOutput *output) const
-{
-    const auto outputs = workspace()->outputs();
-    const auto it = std::ranges::find_if(outputs, [output](LogicalOutput *logical) {
-        return logical->backendOutput() == output;
-    });
-    return it == outputs.end() ? nullptr : *it;
-}
-
 void Compositor::reinitialize()
 {
     // Restart compositing
@@ -299,12 +291,9 @@ void Compositor::start()
 
     createScene();
 
-    const QList<LogicalOutput *> outputs = workspace()->outputs();
-    for (LogicalOutput *output : outputs) {
-        addOutput(output);
-    }
-    connect(workspace(), &Workspace::outputAdded, this, &Compositor::addOutput);
-    connect(workspace(), &Workspace::outputRemoved, this, &Compositor::removeOutput);
+    handleOutputsChanged();
+    connect(workspace(), &Workspace::outputsChanged, this, &Compositor::handleOutputsChanged);
+    connect(kwinApp()->outputBackend(), &OutputBackend::outputRemoved, this, &Compositor::removeOutput);
 
     m_state = State::On;
 
@@ -338,8 +327,8 @@ void Compositor::stop()
         for (Window *window : windows) {
             window->finishCompositing();
         }
-        disconnect(workspace(), &Workspace::outputAdded, this, &Compositor::addOutput);
-        disconnect(workspace(), &Workspace::outputRemoved, this, &Compositor::removeOutput);
+        disconnect(workspace(), &Workspace::outputsChanged, this, &Compositor::handleOutputsChanged);
+        disconnect(kwinApp()->outputBackend(), &OutputBackend::outputRemoved, this, &Compositor::removeOutput);
     }
 
     if (m_backend->compositingType() == OpenGLCompositing) {
@@ -351,7 +340,7 @@ void Compositor::stop()
         return pair.first;
     }) | std::ranges::to<QList>();
     for (RenderLoop *loop : loops) {
-        removeOutput(findLogicalOutput(findOutput(loop)));
+        removeOutput(findOutput(loop));
     }
 
     m_scene.reset();
@@ -373,42 +362,31 @@ static bool isTearingRequested(const Item *item)
     });
 }
 
-static bool checkForBlackBackground(SurfaceItem *background)
+static Rect mapGlobalLogicalToOutputDeviceCoordinates(const RectF &logicalGeometry, LogicalOutput *logicalOutput, BackendOutput *backendOutput)
 {
-    if (!background->buffer()
-        || (!background->buffer()->singlePixelAttributes() && !background->buffer()->shmAttributes())
-        || background->buffer()->size() != QSize(1, 1)) {
-        return false;
-    }
-    const GraphicsBufferView view(background->buffer());
-    if (!view.image()) {
-        return false;
-    }
-    const QRgb rgb = view.image()->pixel(0, 0);
-    const QVector3D encoded(qRed(rgb) / 255.0, qGreen(rgb) / 255.0, qBlue(rgb) / 255.0);
-    const QVector3D nits = background->colorDescription()->mapTo(encoded, ColorDescription(Colorimetry::BT709, TransferFunction(TransferFunction::linear), 100, 0, std::nullopt, std::nullopt), background->renderingIntent());
-    // below 0.1 nits, it shouldn't be noticeable that we replace it with black
-    return nits.lengthSquared() <= (0.1 * 0.1);
+    const Rect localDevice = logicalGeometry.scaled(backendOutput->scale()).rounded();
+    const QPoint scaledOutputPos = (logicalOutput->geometryF().topLeft() * backendOutput->scale()).toPoint();
+    return backendOutput->transform().map(localDevice.translated(backendOutput->deviceOffset() - scaledOutputPos), backendOutput->pixelSize());
 }
 
-static bool prepareDirectScanout(RenderView *view, BackendOutput *output, const std::shared_ptr<OutputFrame> &frame)
+static Rect mapItemToOutputDeviceCoordinates(Item *item, RenderView *view, LogicalOutput *logicalOutput, BackendOutput *backendOutput)
 {
-    if (!view->isVisible() || !view->viewport().intersects(output->geometryF())) {
+    const Rect scaledItemRect = item->mapToView(item->rect(), view).scaled(view->scale()).rounded();
+    const QPoint scaledOutputPos = (logicalOutput->geometryF().topLeft() * view->scale()).toPoint();
+    return backendOutput->transform().map(scaledItemRect.translated(backendOutput->deviceOffset() - scaledOutputPos), backendOutput->pixelSize());
+}
+
+static bool prepareDirectScanout(RenderView *view, LogicalOutput *logicalOutput, BackendOutput *backendOutput, const std::shared_ptr<OutputFrame> &frame)
+{
+    if (!view->isVisible()) {
         return false;
     }
     const auto layer = view->layer();
-    const auto outputLocalRect = view->viewport().translated(-output->geometryF().topLeft());
-    const auto nativeViewport = scaledRect(outputLocalRect, output->scale()).toRect();
-    const bool coversEntireOutput = nativeViewport == QRect(QPoint(), output->pixelSize());
-    // the background of the output can be assumed to be black
-    const auto scanoutCandidates = view->scanoutCandidates(coversEntireOutput ? 2 : 1);
+    const auto scanoutCandidates = view->scanoutCandidates(1);
     if (scanoutCandidates.isEmpty()) {
         layer->setScanoutCandidate(nullptr);
         return false;
     }
-    if (coversEntireOutput && scanoutCandidates.size() == 2 && !checkForBlackBackground(scanoutCandidates.back())) {
-        return false;
-    }
     SurfaceItem *candidate = scanoutCandidates.front();
     SurfaceItemWayland *wayland = qobject_cast<SurfaceItemWayland *>(candidate);
     if (!wayland || !wayland->surface()) {
@@ -429,13 +407,12 @@ static bool prepareDirectScanout(RenderView *view, BackendOutput *output, const
         candidate->setScanoutHint(layer->scanoutDevice(), formats);
         return false;
     }
-    const auto geometry = candidate->mapToView(QRectF(QPointF(0, 0), candidate->size()), view).translated(-output->geometryF().topLeft());
-    layer->setTargetRect(output->transform().map(scaledRect(geometry, output->scale()), output->pixelSize()).toRect());
+    layer->setTargetRect(mapItemToOutputDeviceCoordinates(candidate, view, logicalOutput, backendOutput));
     layer->setEnabled(true);
     layer->setSourceRect(candidate->bufferSourceBox());
     layer->setBufferTransform(candidate->bufferTransform());
-    layer->setOffloadTransform(candidate->bufferTransform().combine(output->transform().inverted()));
-    layer->setColor(candidate->colorDescription(), candidate->renderingIntent(), ColorPipeline::create(candidate->colorDescription(), output->layerBlendingColor(), candidate->renderingIntent()));
+    layer->setOffloadTransform(candidate->bufferTransform().combine(backendOutput->transform().inverted()));
+    layer->setColor(candidate->colorDescription(), candidate->renderingIntent(), ColorPipeline::create(candidate->colorDescription(), backendOutput->layerBlendingColor(), candidate->renderingIntent()));
     const bool ret = layer->importScanoutBuffer(candidate->buffer(), frame);
     if (ret) {
         candidate->resetDamage();
@@ -445,38 +422,44 @@ static bool prepareDirectScanout(RenderView *view, BackendOutput *output, const
     return ret;
 }
 
-static bool prepareRendering(RenderView *view, BackendOutput *output, uint32_t requiredAlphaBits)
+static bool prepareRendering(RenderView *view, LogicalOutput *logicalOutput, BackendOutput *backendOutput, uint32_t requiredAlphaBits)
 {
-    if (!view->isVisible() || !view->viewport().intersects(output->geometryF())) {
+    if (!view->isVisible()) {
         return false;
     }
+    auto nativeRect = mapGlobalLogicalToOutputDeviceCoordinates(view->viewport(), logicalOutput, backendOutput);
+    // we need to render black bars for mirroring,
+    // so add the relevant area to the source and target rect
+    const QSize renderOffset = backendOutput->transform().map(QSize(view->renderOffset().x(), view->renderOffset().y()));
+    nativeRect.adjust(-renderOffset.width(), -renderOffset.height(), renderOffset.width(), renderOffset.height());
+    if ((nativeRect & Rect(QPoint(), backendOutput->modeSize())).isEmpty()) {
+        return false;
+    }
+
     const auto layer = view->layer();
-    const auto outputLocalRect = view->viewport().translated(-output->geometryF().topLeft());
-    const auto nativeRect = output->transform().map(scaledRect(outputLocalRect, output->scale()), output->pixelSize()).toRect();
-    const double reference = output->colorDescription()->referenceLuminance();
-    const double maxOutputLuminance = output->colorDescription()->maxHdrLuminance().value_or(reference);
+    const double reference = backendOutput->colorDescription()->referenceLuminance();
+    const double maxOutputLuminance = backendOutput->colorDescription()->maxHdrLuminance().value_or(reference);
     const double usedMaxLuminance = std::min(view->desiredHdrHeadroom() * reference, maxOutputLuminance);
-    layer->setSourceRect(QRect(QPoint(0, 0), nativeRect.size()));
+    layer->setSourceRect(Rect(QPoint(0, 0), nativeRect.size()));
     layer->setTargetRect(nativeRect);
-    layer->setHotspot(output->transform().map(view->hotspot() * output->scale(), nativeRect.size()));
+    layer->setHotspot(backendOutput->transform().map(view->hotspot() * view->scale(), nativeRect.size()));
     layer->setEnabled(true);
     layer->setOffloadTransform(OutputTransform::Normal);
-    layer->setBufferTransform(output->transform());
-    layer->setColor(output->layerBlendingColor()->withHdrMetadata(reference, usedMaxLuminance), RenderingIntent::AbsoluteColorimetricNoAdaptation, ColorPipeline{});
+    layer->setBufferTransform(backendOutput->transform());
+    layer->setColor(backendOutput->layerBlendingColor()->withHdrMetadata(reference, usedMaxLuminance), RenderingIntent::AbsoluteColorimetricNoAdaptation, ColorPipeline{});
     layer->setRequiredAlphaBits(requiredAlphaBits);
     return layer->preparePresentationTest();
 }
 
-static bool renderLayer(RenderView *view, BackendOutput *output, const std::shared_ptr<OutputFrame> &frame, const QRegion &surfaceDamage)
+static bool renderLayer(RenderView *view, LogicalOutput *logicalOutput, BackendOutput *backendOutput, const std::shared_ptr<OutputFrame> &frame, const Region &surfaceDamage)
 {
     auto beginInfo = view->layer()->beginFrame();
     if (!beginInfo) {
         return false;
     }
     auto &[renderTarget, repaint] = beginInfo.value();
-    const QRect surfaceRect = QRect(QPoint(), renderTarget.transform().map(view->layer()->sourceRect().size().toSize()));
-    const QRegion bufferDamage = surfaceDamage.united(repaint).intersected(surfaceRect);
-    view->paint(renderTarget, bufferDamage);
+    const Region bufferDamage = surfaceDamage.united(repaint).intersected(renderTarget.transformedRect());
+    view->paint(renderTarget, view->renderOffset(), bufferDamage);
     return view->layer()->endFrame(bufferDamage, surfaceDamage, frame.get());
 }
 
@@ -491,12 +474,10 @@ static OutputLayer *findLayer(std::span<OutputLayer *const> layers, OutputLayerT
     return it == layers.end() ? nullptr : *it;
 }
 
-static const bool s_forceSoftwareCursor = environmentVariableBoolValue("KWIN_FORCE_SW_CURSOR").value_or(false);
-
 /**
  * items and layers need to be sorted top to bottom
  */
-static std::unordered_map<SurfaceItem *, OutputLayer *> assignOverlays(RenderView *sceneView, std::span<SurfaceItem *const> underlays, std::span<SurfaceItem *const> overlays, std::span<OutputLayer *const> layers)
+static std::unordered_map<Item *, OutputLayer *> assignOverlays(RenderView *sceneView, std::span<Item *const> underlays, std::span<Item *const> overlays, std::span<OutputLayer *const> layers)
 {
     if (layers.empty() || (underlays.empty() && overlays.empty())) {
         return {};
@@ -505,11 +486,11 @@ static std::unordered_map<SurfaceItem *, OutputLayer *> assignOverlays(RenderVie
     const int primaryZpos = sceneView->layer()->zpos();
     auto layerIt = layers.begin();
     int zpos = (*layerIt)->maxZpos();
-    std::unordered_map<SurfaceItem *, OutputLayer *> ret;
+    std::unordered_map<Item *, OutputLayer *> ret;
     auto overlaysIt = overlays.begin();
     for (; overlaysIt != overlays.end();) {
-        SurfaceItem *item = *overlaysIt;
-        const QRectF sceneRect = item->mapToView(item->rect(), sceneView);
+        Item *item = *overlaysIt;
+        const RectF sceneRect = item->mapToView(item->rect(), sceneView);
         if (sceneRect.contains(sceneView->viewport())) {
             // leave fullscreen direct scanout to the primary plane
             overlaysIt++;
@@ -528,10 +509,24 @@ static std::unordered_map<SurfaceItem *, OutputLayer *> assignOverlays(RenderVie
             // can't use this
             return {};
         }
-        if (!layer->recommendedSizes().isEmpty()) {
+        if (layer->type() == OutputLayerType::CursorOnly && qobject_cast<CursorItem *>(item) == nullptr) {
+            layerIt++;
+            continue;
+        }
+        const auto recommendedSizes = layer->recommendedSizes();
+        if (!recommendedSizes.isEmpty()) {
             // it's likely that sizes other than the recommended ones won't work
-            const QRect deviceRect = scaledRect(sceneRect.translated(-sceneView->viewport().topLeft()), sceneView->scale()).toRect();
-            if (!layer->recommendedSizes().contains(deviceRect.size())) {
+            const bool compositingAllowed = qobject_cast<CursorItem *>(item) != nullptr;
+            const Rect deviceRect = sceneRect.translated(-sceneView->viewport().topLeft()).scaled(sceneView->scale()).rounded();
+            const bool hasFittingSize = std::ranges::any_of(recommendedSizes, [compositingAllowed, deviceRect](const QSize &size) {
+                if (compositingAllowed) {
+                    return deviceRect.size().width() <= size.width()
+                        && deviceRect.size().height() <= size.height();
+                } else {
+                    return deviceRect.size() == size;
+                }
+            });
+            if (!hasFittingSize) {
                 layerIt++;
                 continue;
             }
@@ -556,8 +551,8 @@ static std::unordered_map<SurfaceItem *, OutputLayer *> assignOverlays(RenderVie
     zpos = std::min(primaryZpos - 1, (*layerIt)->maxZpos());
     auto underlaysIt = underlays.begin();
     for (; underlaysIt != underlays.end();) {
-        SurfaceItem *item = *underlaysIt;
-        const QRectF sceneRect = item->mapToView(item->rect(), sceneView);
+        Item *item = *underlaysIt;
+        const RectF sceneRect = item->mapToView(item->rect(), sceneView);
         if (sceneRect.contains(sceneView->viewport())) {
             // leave fullscreen direct scanout to the primary plane
             underlaysIt++;
@@ -572,10 +567,24 @@ static std::unordered_map<SurfaceItem *, OutputLayer *> assignOverlays(RenderVie
             layerIt++;
             continue;
         }
-        if (!layer->recommendedSizes().isEmpty()) {
+        if (layer->type() == OutputLayerType::CursorOnly && qobject_cast<CursorItem *>(item) == nullptr) {
+            layerIt++;
+            continue;
+        }
+        const auto recommendedSizes = layer->recommendedSizes();
+        if (!recommendedSizes.isEmpty()) {
             // it's likely that sizes other than the recommended ones won't work
-            const QRect deviceRect = scaledRect(sceneRect.translated(-sceneView->viewport().topLeft()), sceneView->scale()).toRect();
-            if (!layer->recommendedSizes().contains(deviceRect.size())) {
+            const bool compositingAllowed = qobject_cast<CursorItem *>(item) != nullptr;
+            const Rect deviceRect = sceneRect.translated(-sceneView->viewport().topLeft()).scaled(sceneView->scale()).rounded();
+            const bool hasFittingSize = std::ranges::any_of(recommendedSizes, [compositingAllowed, deviceRect](const QSize &size) {
+                if (compositingAllowed) {
+                    return deviceRect.size().width() <= size.width()
+                        && deviceRect.size().height() <= size.height();
+                } else {
+                    return deviceRect.size() == size;
+                }
+            });
+            if (!hasFittingSize) {
                 layerIt++;
                 continue;
             }
@@ -608,7 +617,7 @@ void Compositor::composite(RenderLoop *renderLoop)
     }
 
     BackendOutput *output = findOutput(renderLoop);
-    LogicalOutput *logical = findLogicalOutput(output);
+    LogicalOutput *logicalOutput = workspace()->findOutput(output);
     const auto primaryView = m_primaryViews[renderLoop].get();
     fTraceDuration("Paint (", output->name(), ")");
 
@@ -628,7 +637,17 @@ void Compositor::composite(RenderLoop *renderLoop)
         || (!output->highDynamicRange() && output->brightnessDevice() && output->brightnessDevice()->brightnessSteps() < 5)) {
         frame->setBrightness(output->brightnessSetting() * output->dimming());
     } else {
-        constexpr double changePerSecond = 3;
+        // animate much slower for automatic brightness
+        double changePerSecond = 3;
+        if (output->lastBrightnessAdjustmentReason() == BackendOutput::BrightnessReason::AutomaticBrightness) {
+            if (output->brightnessSetting() < output->currentBrightness()) {
+                // brightness should be reduced slowly, or it'll be annoying
+                changePerSecond = 0.1;
+            } else {
+                // but increased more quickly, so that you can still read your screen
+                changePerSecond = 0.5;
+            }
+        }
         const double maxChangePerFrame = changePerSecond * 1'000.0 / renderLoop->refreshRate();
         // brightness perception is non-linear, gamma 2.2 encoding *roughly* represents that
         const double current = std::pow(*output->currentBrightness(), 1.0 / 2.2);
@@ -636,10 +655,10 @@ void Compositor::composite(RenderLoop *renderLoop)
     }
 
     Window *const activeWindow = workspace()->activeWindow();
-    SurfaceItem *const activeFullscreenItem = activeWindow && activeWindow->isFullScreen() && activeWindow->frameGeometry().intersects(output->geometryF()) ? activeWindow->surfaceItem() : nullptr;
+    SurfaceItem *const activeFullscreenItem = activeWindow && activeWindow->isFullScreen() && activeWindow->frameGeometry().intersects(primaryView->viewport()) ? activeWindow->surfaceItem() : nullptr;
     frame->setContentType(activeWindow && activeFullscreenItem ? activeFullscreenItem->contentType() : ContentType::None);
 
-    const bool wantsAdaptiveSync = activeWindow && activeWindow->frameGeometry().intersects(output->geometryF()) && activeWindow->wantsAdaptiveSync();
+    const bool wantsAdaptiveSync = activeWindow && activeWindow->frameGeometry().intersects(primaryView->viewport()) && activeWindow->wantsAdaptiveSync();
     const bool vrr = (output->capabilities() & BackendOutput::Capability::Vrr) && (output->vrrPolicy() == VrrPolicy::Always || (output->vrrPolicy() == VrrPolicy::Automatic && wantsAdaptiveSync));
     const bool tearing = (output->capabilities() & BackendOutput::Capability::Tearing) && options->allowTearing() && activeFullscreenItem && activeWindow->wantsTearing(isTearingRequested(activeFullscreenItem));
     if (vrr) {
@@ -655,7 +674,7 @@ void Compositor::composite(RenderLoop *renderLoop)
         bool directScanout = false;
         bool directScanoutOnly = false;
         bool highPriority = false;
-        QRegion surfaceDamage;
+        Region surfaceDamage;
         uint32_t requiredAlphaBits;
     };
     QList<LayerData> layers;
@@ -666,7 +685,7 @@ void Compositor::composite(RenderLoop *renderLoop)
         .directScanout = false,
         .directScanoutOnly = false,
         .highPriority = false,
-        .surfaceDamage = QRegion{},
+        .surfaceDamage = Region{},
         .requiredAlphaBits = 0,
     });
 
@@ -697,24 +716,38 @@ void Compositor::composite(RenderLoop *renderLoop)
     // the primary output layer is currently always used for the main content
     unusedOutputLayers.removeOne(primaryView->layer());
 
-    OutputLayer *cursorLayer = nullptr;
-    Item *cursorItem = m_scene->cursorItem();
-    if (!s_forceSoftwareCursor
-        && !m_brokenCursors.contains(renderLoop)
-        && cursorItem->isVisible()
-        && cursorItem->mapToView(cursorItem->boundingRect(), primaryView).intersects(output->geometryF())) {
-        cursorLayer = findLayer(unusedOutputLayers, OutputLayerType::CursorOnly, primaryView->layer()->zpos() + 1);
-        if (!cursorLayer) {
-            cursorLayer = findLayer(unusedOutputLayers, OutputLayerType::EfficientOverlay, primaryView->layer()->zpos() + 1);
-        }
-        if (!cursorLayer) {
-            cursorLayer = findLayer(unusedOutputLayers, OutputLayerType::GenericLayer, primaryView->layer()->zpos() + 1);
-        }
-        if (cursorLayer) {
-            auto &view = m_overlayViews[renderLoop][cursorLayer];
-            if (!view || view->item() != cursorItem) {
-                view = std::make_unique<ItemTreeView>(primaryView, cursorItem, logical, cursorLayer);
-                connect(cursorLayer, &OutputLayer::repaintScheduled, view.get(), [output, cursorView = view.get()]() {
+    const bool overlaysAllowed = m_allowOverlaysEnv.value_or(!output->overlayLayersLikelyBroken() && PROJECT_VERSION_PATCH >= 80);
+    QList<OutputLayer *> specialLayers = unusedOutputLayers | std::views::filter([this, renderLoop, overlaysAllowed](OutputLayer *layer) {
+        return layer->type() != OutputLayerType::Primary
+            && (!m_brokenCursors.contains(renderLoop) || layer->type() != OutputLayerType::CursorOnly)
+            && (overlaysAllowed || layer->type() != OutputLayerType::GenericLayer);
+    }) | std::ranges::to<QList>();
+    std::ranges::sort(specialLayers, [](OutputLayer *left, OutputLayer *right) {
+        return left->maxZpos() > right->maxZpos();
+    });
+    const size_t maxOverlayCount = std::ranges::count_if(specialLayers, [primaryView](OutputLayer *layer) {
+        return layer->maxZpos() > primaryView->layer()->zpos();
+    });
+    const size_t maxUnderlayCount = std::ranges::count_if(specialLayers, [primaryView](OutputLayer *layer) {
+        return layer->minZpos() < primaryView->layer()->zpos();
+    });
+    const auto [overlayCandidates, underlayCandidates] = m_scene->overlayCandidates(specialLayers.size(), maxOverlayCount, maxUnderlayCount);
+    auto overlayAssignments = assignOverlays(primaryView, underlayCandidates, overlayCandidates, specialLayers);
+    if (overlayAssignments.empty()) {
+        // the cursor is important, so try again without other over/underlays
+        const auto cursorOnly = overlayCandidates | std::views::filter([](Item *item) {
+            return qobject_cast<CursorItem *>(item) != nullptr;
+        }) | std::ranges::to<QList>();
+        overlayAssignments = assignOverlays(primaryView, {}, cursorOnly, specialLayers);
+    }
+    for (const auto &[item, layer] : overlayAssignments) {
+        const bool isCursor = qobject_cast<CursorItem *>(item) != nullptr;
+        auto &view = m_overlayViews[output->renderLoop()][layer];
+        if (!view || view->item() != item) {
+            if (isCursor) {
+                // special handling for the cursor
+                view = std::make_unique<ItemTreeView>(primaryView, item, logicalOutput, output, layer);
+                connect(layer, &OutputLayer::repaintScheduled, view.get(), [logicalOutput, output, cursorView = view.get()]() {
                     // this just deals with moving the plane asynchronously, for improved latency.
                     // enabling, disabling and updating the cursor image still happen in composite()
                     const auto outputLayer = cursorView->layer();
@@ -734,9 +767,7 @@ void Compositor::composite(RenderLoop *renderLoop)
                         }).value_or(30);
                         maxVrrCursorDelay = std::chrono::nanoseconds(1'000'000'000) / std::max(effectiveMinRate, 30u);
                     }
-                    const QRectF outputLocalRect = output->mapFromGlobal(cursorView->viewport());
-                    const QRectF nativeCursorRect = output->transform().map(QRectF(outputLocalRect.topLeft() * output->scale(), outputLayer->targetRect().size()), output->pixelSize());
-                    outputLayer->setTargetRect(QRect(nativeCursorRect.topLeft().toPoint(), outputLayer->targetRect().size()));
+                    outputLayer->setTargetRect(mapGlobalLogicalToOutputDeviceCoordinates(cursorView->viewport(), logicalOutput, output));
                     outputLayer->setEnabled(true);
                     if (output->presentAsync(outputLayer, maxVrrCursorDelay)) {
                         // prevent composite() from also pushing an update with the cursor layer
@@ -744,52 +775,18 @@ void Compositor::composite(RenderLoop *renderLoop)
                         outputLayer->resetRepaints();
                     }
                 });
+            } else {
+                view = std::make_unique<ItemView>(primaryView, item, logicalOutput, output, layer);
             }
-            view->prePaint();
-            layers.push_back(LayerData{
-                .view = view.get(),
-                .directScanout = false,
-                .directScanoutOnly = false,
-                .highPriority = true,
-                .surfaceDamage = QRegion{},
-                .requiredAlphaBits = 8,
-            });
-            cursorLayer->setZpos(cursorLayer->maxZpos());
-            unusedOutputLayers.removeOne(cursorLayer);
-        }
-    }
-
-    QList<OutputLayer *> specialLayers = unusedOutputLayers | std::views::filter([cursorLayer](OutputLayer *layer) {
-        return layer->type() != OutputLayerType::Primary
-            && (!cursorLayer || layer->minZpos() < cursorLayer->zpos());
-    }) | std::ranges::to<QList>();
-    std::ranges::sort(specialLayers, [](OutputLayer *left, OutputLayer *right) {
-        return left->maxZpos() > right->maxZpos();
-    });
-    const size_t maxOverlayCount = std::ranges::count_if(specialLayers, [primaryView](OutputLayer *layer) {
-        return layer->maxZpos() > primaryView->layer()->zpos();
-    });
-    const size_t maxUnderlayCount = std::ranges::count_if(specialLayers, [primaryView](OutputLayer *layer) {
-        return layer->minZpos() < primaryView->layer()->zpos();
-    });
-    const auto [overlayCandidates, underlayCandidates] = m_scene->overlayCandidates(specialLayers.size(), maxOverlayCount, maxUnderlayCount);
-    std::unordered_map<SurfaceItem *, OutputLayer *> overlayAssignments;
-    if (m_allowOverlaysEnv.value_or(!output->overlayLayersLikelyBroken() && PROJECT_VERSION_PATCH >= 80)) {
-        overlayAssignments = assignOverlays(primaryView, underlayCandidates, overlayCandidates, specialLayers);
-    }
-    for (const auto &[item, layer] : overlayAssignments) {
-        auto &view = m_overlayViews[output->renderLoop()][layer];
-        if (!view || view->item() != item) {
-            view = std::make_unique<ItemView>(primaryView, item, logical, layer);
         }
         view->prePaint();
         layers.push_back(LayerData{
             .view = view.get(),
-            .directScanout = true,
-            .directScanoutOnly = true,
-            .highPriority = false,
-            .surfaceDamage = QRegion(),
-            .requiredAlphaBits = 0,
+            .directScanout = !isCursor,
+            .directScanoutOnly = !isCursor,
+            .highPriority = isCursor,
+            .surfaceDamage = Region(),
+            .requiredAlphaBits = isCursor ? 8u : 0u,
         });
         unusedOutputLayers.removeOne(layer);
         if (layer->zpos() < primaryView->layer()->zpos()) {
@@ -814,9 +811,9 @@ void Compositor::composite(RenderLoop *renderLoop)
 
     // update all of them for the ideal configuration
     for (auto &layer : layers) {
-        if (prepareDirectScanout(layer.view, output, frame)) {
+        if (prepareDirectScanout(layer.view, logicalOutput, output, frame)) {
             layer.directScanout = true;
-        } else if (!layer.directScanoutOnly && prepareRendering(layer.view, output, layer.requiredAlphaBits)) {
+        } else if (!layer.directScanoutOnly && prepareRendering(layer.view, logicalOutput, output, layer.requiredAlphaBits)) {
             layer.directScanout = false;
         } else {
             layer.view->layer()->setEnabled(false);
@@ -830,7 +827,7 @@ void Compositor::composite(RenderLoop *renderLoop)
         bool primaryFailure = false;
         auto &primary = layers.front();
         if (primary.directScanout) {
-            if (prepareRendering(primary.view, output, primary.requiredAlphaBits)) {
+            if (prepareRendering(primary.view, logicalOutput, output, primary.requiredAlphaBits)) {
                 primary.directScanout = false;
                 result = output->testPresentation(frame);
             } else {
@@ -882,7 +879,7 @@ void Compositor::composite(RenderLoop *renderLoop)
             layer.surfaceDamage |= layer.view->layer()->deviceRepaints();
             layer.view->layer()->resetRepaints();
             if (layer.view->layer()->isEnabled() && !layer.directScanout) {
-                result &= renderLayer(layer.view, output, frame, layer.surfaceDamage);
+                result &= renderLayer(layer.view, logicalOutput, output, frame, layer.surfaceDamage);
                 if (!result) {
                     qCWarning(KWIN_CORE, "Rendering a layer failed!");
                     break;
@@ -915,8 +912,8 @@ void Compositor::composite(RenderLoop *renderLoop)
                 layer.view->setExclusive(false);
             }
             // re-render without direct scanout
-            if (prepareRendering(primary.view, output, primary.requiredAlphaBits)
-                && renderLayer(primary.view, output, frame, primary.surfaceDamage)) {
+            if (prepareRendering(primary.view, logicalOutput, output, primary.requiredAlphaBits)
+                && renderLayer(primary.view, logicalOutput, output, frame, primary.surfaceDamage)) {
                 result = output->present(toUpdate, frame);
             } else {
                 qCWarning(KWIN_CORE, "Rendering the primary layer failed!");
@@ -928,8 +925,8 @@ void Compositor::composite(RenderLoop *renderLoop)
             // try again even without the cursor layer
             layers[1].view->layer()->setEnabled(false);
             layers[1].view->setExclusive(false);
-            if (prepareRendering(primary.view, output, primary.requiredAlphaBits)
-                && renderLayer(primary.view, output, frame, infiniteRegion())) {
+            if (prepareRendering(primary.view, logicalOutput, output, primary.requiredAlphaBits)
+                && renderLayer(primary.view, logicalOutput, output, frame, Region::infinite())) {
                 result = output->present(toUpdate, frame);
                 if (result) {
                     // disabling the cursor layer helped... so disable it permanently,
@@ -945,11 +942,9 @@ void Compositor::composite(RenderLoop *renderLoop)
         }
     }
 
+    m_scene->frame(primaryView, frame.get());
     for (auto &layer : layers) {
         layer.view->postPaint();
-        if (layer.view->layer()->isEnabled()) {
-            layer.view->frame(frame.get());
-        }
     }
 
     // the layers have to stay valid until after postPaint, so this needs to happen after it
@@ -975,9 +970,21 @@ void Compositor::composite(RenderLoop *renderLoop)
     }
 }
 
-void Compositor::addOutput(LogicalOutput *logical)
+void Compositor::handleOutputsChanged()
+{
+    for (auto &[loop, layer] : m_primaryViews) {
+        disconnect(loop, &RenderLoop::frameRequested, this, &Compositor::handleFrameRequested);
+    }
+    m_overlayViews.clear();
+    m_primaryViews.clear();
+    const auto outputs = kwinApp()->outputBackend()->outputs();
+    for (BackendOutput *output : outputs | std::views::filter(&BackendOutput::isEnabled)) {
+        addOutput(output);
+    }
+}
+
+void Compositor::addOutput(BackendOutput *output)
 {
-    BackendOutput *output = logical->backendOutput();
     if (output->isPlaceholder()) {
         return;
     }
@@ -988,9 +995,8 @@ void Compositor::addOutput(LogicalOutput *logical)
     });
 }
 
-void Compositor::removeOutput(LogicalOutput *logical)
+void Compositor::removeOutput(BackendOutput *output)
 {
-    BackendOutput *output = logical->backendOutput();
     if (output->isPlaceholder()) {
         return;
     }
@@ -1003,7 +1009,7 @@ void Compositor::removeOutput(LogicalOutput *logical)
 
 void Compositor::assignOutputLayers(BackendOutput *output)
 {
-    LogicalOutput *logical = findLogicalOutput(output);
+    LogicalOutput *logical = workspace()->findOutput(output);
     Q_ASSERT(logical);
     const auto layers = m_backend->compatibleOutputLayers(output);
     const auto primaryLayer = findLayer(layers, OutputLayerType::Primary, std::nullopt);
@@ -1012,15 +1018,22 @@ void Compositor::assignOutputLayers(BackendOutput *output)
     if (sceneView) {
         sceneView->setLayer(primaryLayer);
     } else {
-        sceneView = std::make_unique<SceneView>(m_scene.get(), logical, primaryLayer);
-        sceneView->setViewport(output->geometryF());
+        sceneView = std::make_unique<SceneView>(m_scene.get(), logical, output, primaryLayer);
         sceneView->setScale(output->scale());
-        connect(output, &BackendOutput::geometryChanged, sceneView.get(), [output, view = sceneView.get()]() {
-            view->setViewport(output->geometryF());
-        });
-        connect(output, &BackendOutput::scaleChanged, sceneView.get(), [output, view = sceneView.get()]() {
+        sceneView->setRenderOffset(output->deviceOffset());
+        const auto updateViewport = [view = sceneView.get(), logical, output]() {
+            // this matches how the renderer snaps elements to the pixel grid
+            const Rect scaled = logical->geometryF().scaled(output->scale()).rounded();
+            view->setViewport(scaled.scaled(1.0 / output->scale()));
+        };
+        updateViewport();
+        connect(logical, &LogicalOutput::geometryChanged, sceneView.get(), updateViewport);
+        connect(output, &BackendOutput::scaleChanged, sceneView.get(), [view = sceneView.get(), output]() {
             view->setScale(output->scale());
         });
+        connect(output, &BackendOutput::deviceOffsetChanged, sceneView.get(), [view = sceneView.get(), output]() {
+            view->setRenderOffset(output->deviceOffset());
+        });
     }
     // will be re-assigned in the next composite() pass
     m_overlayViews.erase(output->renderLoop());
