@@ -442,7 +442,7 @@ double WorkspaceScene::desiredHdrHeadroom() const
 
 void WorkspaceScene::frame(SceneView *delegate, OutputFrame *frame)
 {
-    LogicalOutput *logicalOutput = delegate->output();
+    LogicalOutput *logicalOutput = delegate->logicalOutput();
     const auto frameTime = std::chrono::duration_cast<std::chrono::milliseconds>(logicalOutput->backendOutput()->renderLoop()->lastPresentationTimestamp());
     m_containerItem->framePainted(delegate, logicalOutput, frame, frameTime);
     if (m_overlayItem) {
@@ -453,7 +453,7 @@ void WorkspaceScene::frame(SceneView *delegate, OutputFrame *frame)
 void WorkspaceScene::prePaint(SceneView *delegate)
 {
     painted_delegate = delegate;
-    painted_screen = painted_delegate->output();
+    painted_screen = painted_delegate->logicalOutput();
 
     createStackingOrder();
 
@@ -477,7 +477,7 @@ void WorkspaceScene::prePaint(SceneView *delegate)
     Q_EMIT preFrameRender();
 
     effects->prePaintScreen(prePaintData, m_expectedPresentTimestamp);
-    m_paintContext.deviceDamage = painted_delegate->mapToDeviceCoordinatesAligned(prePaintData.paint);
+    m_paintContext.deviceDamage = painted_delegate->mapToDeviceCoordinatesAligned(prePaintData.paint) & painted_delegate->deviceRect();
     m_paintContext.mask = prePaintData.mask;
     m_paintContext.phase2Data.clear();
 
@@ -563,7 +563,7 @@ void WorkspaceScene::preparePaintSimpleScreen()
         effects->prePaintWindow(painted_delegate, windowItem->effectWindow(), data, m_expectedPresentTimestamp);
         m_paintContext.phase2Data.append(Phase2Data{
             .item = windowItem,
-            .deviceRegion = data.devicePaint,
+            .deviceRegion = data.devicePaint & painted_delegate->deviceRect(),
             .deviceOpaque = data.deviceOpaque,
             .mask = data.mask,
         });
@@ -574,8 +574,8 @@ QRegion WorkspaceScene::collectDamage()
 {
     if (m_paintContext.mask & (PAINT_SCREEN_TRANSFORMED | PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS)) {
         resetRepaintsHelper(m_overlayItem.get(), painted_delegate);
-        m_paintContext.deviceDamage = infiniteRegion();
-        return infiniteRegion();
+        m_paintContext.deviceDamage = painted_delegate->deviceRect();
+        return m_paintContext.deviceDamage;
     } else {
         // Perform an occlusion cull pass, remove surface damage occluded by opaque windows.
         QRegion opaque;
@@ -583,14 +583,17 @@ QRegion WorkspaceScene::collectDamage()
             auto &paintData = m_paintContext.phase2Data[i];
             accumulateRepaints(paintData.item, painted_delegate, &paintData.deviceRegion);
             m_paintContext.deviceDamage += paintData.deviceRegion - opaque;
-            if (!(paintData.mask & (PAINT_WINDOW_TRANSLUCENT | PAINT_WINDOW_TRANSFORMED))) {
+            // TODO change effects API, so occlusion culling is per item, rather than per window
+            const bool canCover = painted_delegate->shouldRenderItem(paintData.item->surfaceItem())
+                || painted_delegate->shouldRenderHole(paintData.item->surfaceItem());
+            if (!(paintData.mask & (PAINT_WINDOW_TRANSLUCENT | PAINT_WINDOW_TRANSFORMED)) && canCover) {
                 opaque += paintData.deviceOpaque;
             }
         }
 
         accumulateRepaints(m_overlayItem.get(), painted_delegate, &m_paintContext.deviceDamage);
 
-        return m_paintContext.deviceDamage;
+        return m_paintContext.deviceDamage & painted_delegate->deviceRect();
     }
 }
 
@@ -603,9 +606,9 @@ void WorkspaceScene::postPaint()
     clearStackingOrder();
 }
 
-void WorkspaceScene::paint(const RenderTarget &renderTarget, const QRegion &deviceRegion)
+void WorkspaceScene::paint(const RenderTarget &renderTarget, const QPoint &deviceOffset, const QRegion &deviceRegion)
 {
-    RenderViewport viewport(painted_delegate->viewport(), painted_delegate->scale(), renderTarget);
+    RenderViewport viewport(painted_delegate->viewport(), painted_delegate->scale(), renderTarget, deviceOffset);
 
     m_renderer->beginFrame(renderTarget, viewport);
 
@@ -665,12 +668,15 @@ void WorkspaceScene::paintSimpleScreen(const RenderTarget &renderTarget, const R
     QRegion visible = deviceRegion;
     for (int i = m_paintContext.phase2Data.size() - 1; i >= 0; --i) {
         Phase2Data *data = &m_paintContext.phase2Data[i];
-        data->deviceRegion = visible;
+        data->deviceRegion = visible & viewport.deviceRect();
 
         if (!(data->mask & PAINT_WINDOW_TRANSFORMED)) {
             data->deviceRegion &= viewport.mapToDeviceCoordinatesAligned(data->item->mapToScene(data->item->boundingRect()));
 
-            if (!(data->mask & PAINT_WINDOW_TRANSLUCENT)) {
+            // TODO change effects API, so occlusion culling is per item, rather than per window
+            const bool canCover = painted_delegate->shouldRenderItem(data->item->surfaceItem())
+                || painted_delegate->shouldRenderHole(data->item->surfaceItem());
+            if (!(data->mask & PAINT_WINDOW_TRANSLUCENT) && canCover) {
                 visible -= data->deviceOpaque;
             }
         }
