@@ -73,22 +73,6 @@ static QRectF scriptValueToQRectF(const QJSValue &value)
                   value.property(QStringLiteral("height")).toNumber());
 }
 
-static KWin::Rect scriptValueToRect(const QJSValue &value)
-{
-    return KWin::Rect(value.property(QStringLiteral("x")).toInt(),
-                      value.property(QStringLiteral("y")).toInt(),
-                      value.property(QStringLiteral("width")).toInt(),
-                      value.property(QStringLiteral("height")).toInt());
-}
-
-static KWin::RectF scriptValueToRectF(const QJSValue &value)
-{
-    return KWin::RectF(value.property(QStringLiteral("x")).toNumber(),
-                       value.property(QStringLiteral("y")).toNumber(),
-                       value.property(QStringLiteral("width")).toNumber(),
-                       value.property(QStringLiteral("height")).toNumber());
-}
-
 static QPoint scriptValueToPoint(const QJSValue &value)
 {
     return QPoint(value.property(QStringLiteral("x")).toInt(),
@@ -173,13 +157,6 @@ KWin::Script::Script(int id, QString scriptName, QString pluginName, QObject *pa
     if (!QMetaType::hasRegisteredConverterFunction<QJSValue, QSizeF>()) {
         QMetaType::registerConverter<QJSValue, QSizeF>(scriptValueToSizeF);
     }
-
-    if (!QMetaType::hasRegisteredConverterFunction<QJSValue, Rect>()) {
-        QMetaType::registerConverter<QJSValue, Rect>(scriptValueToRect);
-    }
-    if (!QMetaType::hasRegisteredConverterFunction<QJSValue, RectF>()) {
-        QMetaType::registerConverter<QJSValue, RectF>(scriptValueToRectF);
-    }
 }
 
 KWin::Script::~Script()
@@ -654,6 +631,51 @@ KWin::Scripting::Scripting(QObject *parent)
     , m_declarativeScriptSharedContext(new QQmlContext(m_qmlEngine, this))
     , m_workspaceWrapper(new QtScriptWorkspaceWrapper(this))
 {
+    // For plain JavaScript extensions. There's no Rect factory function, we accept any object
+    // with x, y, width, and height properties as rects.
+    if (!QMetaType::hasRegisteredConverterFunction<QJSValue, Rect>()) {
+        QMetaType::registerConverter<QJSValue, Rect>([](const QJSValue &value) {
+            return Rect(value.property(QStringLiteral("x")).toInt(),
+                        value.property(QStringLiteral("y")).toInt(),
+                        value.property(QStringLiteral("width")).toInt(),
+                        value.property(QStringLiteral("height")).toInt());
+        });
+    }
+
+    if (!QMetaType::hasRegisteredConverterFunction<QJSValue, RectF>()) {
+        QMetaType::registerConverter<QJSValue, RectF>([](const QJSValue &value) {
+            return RectF(value.property(QStringLiteral("x")).toNumber(),
+                         value.property(QStringLiteral("y")).toNumber(),
+                         value.property(QStringLiteral("width")).toNumber(),
+                         value.property(QStringLiteral("height")).toNumber());
+        });
+    }
+
+    // For QML extensions.
+    if (!QMetaType::hasRegisteredConverterFunction<QRect, Rect>()) {
+        QMetaType::registerConverter<QRect, Rect>([](const QRect &rect) {
+            return Rect(rect.x(), rect.y(), rect.width(), rect.height());
+        });
+    }
+
+    if (!QMetaType::hasRegisteredConverterFunction<Rect, QRect>()) {
+        QMetaType::registerConverter<Rect, QRect>([](const Rect &rect) {
+            return QRect(rect.x(), rect.y(), rect.width(), rect.height());
+        });
+    }
+
+    if (!QMetaType::hasRegisteredConverterFunction<QRectF, RectF>()) {
+        QMetaType::registerConverter<QRectF, RectF>([](const QRectF &rect) {
+            return RectF(rect.x(), rect.y(), rect.width(), rect.height());
+        });
+    }
+
+    if (!QMetaType::hasRegisteredConverterFunction<RectF, QRectF>()) {
+        QMetaType::registerConverter<RectF, QRectF>([](const RectF &rect) {
+            return QRectF(rect.x(), rect.y(), rect.width(), rect.height());
+        });
+    }
+
     m_qmlEngine->setProperty("_kirigamiTheme", QStringLiteral("KirigamiPlasmaStyle"));
     m_qmlEngine->rootContext()->setContextObject(new KLocalizedQmlContext(m_qmlEngine));
     init();
