@@ -77,6 +77,9 @@ static uint32_t kwinCapabilitiesToOutputDeviceCapabilities(BackendOutput::Capabi
     if (caps & BackendOutput::Capability::SharpnessControl) {
         ret |= QtWaylandServer::kde_output_device_v2::capability_sharpness;
     }
+    if (caps & BackendOutput::Capability::CustomModes) {
+        ret |= QtWaylandServer::kde_output_device_v2::capability_custom_modes;
+    }
     return ret;
 }
 
@@ -153,7 +156,7 @@ public:
     QString m_name;
     subpixel m_subPixel = subpixel_unknown;
     transform m_transform = transform_normal;
-    QList<OutputDeviceModeV2Interface *> m_modes;
+    std::vector<std::unique_ptr<OutputDeviceModeV2Interface>> m_modes;
     OutputDeviceModeV2Interface *m_currentMode = nullptr;
     QByteArray m_edid;
     bool m_enabled = true;
@@ -207,7 +210,7 @@ public:
     Resource *createResource(OutputDeviceV2InterfacePrivate::Resource *output);
     Resource *findResource(OutputDeviceV2InterfacePrivate::Resource *output) const;
 
-    void bindResource(wl_resource *resource);
+    void bindResource(Resource *resource);
 
     static OutputDeviceModeV2InterfacePrivate *get(OutputDeviceModeV2Interface *mode)
     {
@@ -218,7 +221,7 @@ public:
     std::weak_ptr<OutputMode> m_handle;
     QSize m_size;
     int m_refreshRate = 60000;
-    bool m_preferred = false;
+    OutputMode::Flags m_flags;
 
 protected:
     Resource *kde_output_device_mode_v2_allocate() override;
@@ -283,9 +286,9 @@ OutputDeviceV2Interface::OutputDeviceV2Interface(Display *display, BackendOutput
     updateSharpness();
     updatePriority();
 
-    connect(handle, &BackendOutput::geometryChanged,
+    connect(handle, &BackendOutput::positionChanged,
             this, &OutputDeviceV2Interface::updateGlobalPosition);
-    connect(handle, &BackendOutput::scaleChanged,
+    connect(handle, &BackendOutput::scaleSettingChanged,
             this, &OutputDeviceV2Interface::updateScale);
     connect(handle, &BackendOutput::enabledChanged,
             this, &OutputDeviceV2Interface::updateEnabled);
@@ -336,7 +339,6 @@ OutputDeviceV2Interface::OutputDeviceV2Interface(Display *display, BackendOutput
 
 OutputDeviceV2Interface::~OutputDeviceV2Interface()
 {
-    d->globalRemove();
 }
 
 void OutputDeviceV2Interface::remove()
@@ -352,6 +354,8 @@ void OutputDeviceV2Interface::remove()
         displayPrivate->outputdevicesV2.removeOne(this);
     }
 
+    // NOTE that output modes can only be deleted after the global remove,
+    // otherwise the client temporarily has an output without any modes
     d->globalRemove();
 }
 
@@ -382,8 +386,8 @@ void OutputDeviceV2InterfacePrivate::kde_output_device_v2_bind_resource(Resource
     sendName(resource);
     sendSerialNumber(resource);
 
-    for (OutputDeviceModeV2Interface *mode : std::as_const(m_modes)) {
-        sendNewMode(resource, mode);
+    for (const auto &mode : m_modes) {
+        sendNewMode(resource, mode.get());
     }
     sendCurrentMode(resource);
     sendUuid(resource);
@@ -422,7 +426,7 @@ wl_resource *OutputDeviceV2InterfacePrivate::sendNewMode(Resource *resource, Out
 
     send_mode(resource->handle, modeResource->handle);
 
-    privateMode->bindResource(modeResource->handle);
+    privateMode->bindResource(modeResource);
 
     return modeResource->handle;
 }
@@ -652,7 +656,7 @@ void OutputDeviceV2Interface::updatePhysicalSize()
 
 void OutputDeviceV2Interface::updateGlobalPosition()
 {
-    const QPoint arg = d->m_handle->geometry().topLeft();
+    const QPoint arg = d->m_handle->position();
     if (d->m_globalPosition == arg) {
         return;
     }
@@ -705,7 +709,7 @@ void OutputDeviceV2Interface::updateTransform()
 
 void OutputDeviceV2Interface::updateScale()
 {
-    const qreal scale = d->m_handle->scale();
+    const qreal scale = d->m_handle->scaleSetting();
     if (qFuzzyCompare(d->m_scale, scale)) {
         return;
     }
@@ -719,16 +723,15 @@ void OutputDeviceV2Interface::updateScale()
 
 void OutputDeviceV2Interface::updateModes()
 {
-    const auto oldModes = d->m_modes;
-    d->m_modes.clear();
+    auto oldModes = std::move(d->m_modes);
     d->m_currentMode = nullptr;
 
     const auto clientResources = d->resourceMap();
     const auto nativeModes = d->m_handle->modes();
 
     for (const std::shared_ptr<OutputMode> &mode : nativeModes) {
-        OutputDeviceModeV2Interface *deviceMode = new OutputDeviceModeV2Interface(mode, this);
-        d->m_modes.append(deviceMode);
+        d->m_modes.push_back(std::make_unique<OutputDeviceModeV2Interface>(mode));
+        OutputDeviceModeV2Interface *deviceMode = d->m_modes.back().get();
 
         if (d->m_handle->currentMode() == mode) {
             d->m_currentMode = deviceMode;
@@ -743,17 +746,17 @@ void OutputDeviceV2Interface::updateModes()
         d->sendCurrentMode(resource);
     }
 
-    qDeleteAll(oldModes.crbegin(), oldModes.crend());
-
+    // make sure old modes are removed before the done event
+    oldModes.clear();
     scheduleDone();
 }
 
 void OutputDeviceV2Interface::updateCurrentMode()
 {
-    for (OutputDeviceModeV2Interface *mode : std::as_const(d->m_modes)) {
+    for (const auto &mode : d->m_modes) {
         if (mode->handle().lock() == d->m_handle->currentMode()) {
-            if (d->m_currentMode != mode) {
-                d->m_currentMode = mode;
+            if (d->m_currentMode != mode.get()) {
+                d->m_currentMode = mode.get();
                 const auto clientResources = d->resourceMap();
                 for (auto resource : clientResources) {
                     d->sendCurrentMode(resource);
@@ -1117,13 +1120,12 @@ OutputDeviceModeV2InterfacePrivate::OutputDeviceModeV2InterfacePrivate(OutputDev
     , m_handle(handle)
     , m_size(handle->size())
     , m_refreshRate(handle->refreshRate())
-    , m_preferred(handle->flags() & OutputMode::Flag::Preferred)
+    , m_flags(handle->flags())
 {
 }
 
-OutputDeviceModeV2Interface::OutputDeviceModeV2Interface(std::shared_ptr<OutputMode> handle, QObject *parent)
-    : QObject(parent)
-    , d(new OutputDeviceModeV2InterfacePrivate(this, handle))
+OutputDeviceModeV2Interface::OutputDeviceModeV2Interface(std::shared_ptr<OutputMode> handle)
+    : d(new OutputDeviceModeV2InterfacePrivate(this, handle))
 {
 }
 
@@ -1166,13 +1168,23 @@ std::weak_ptr<OutputMode> OutputDeviceModeV2Interface::handle() const
     return d->m_handle;
 }
 
-void OutputDeviceModeV2InterfacePrivate::bindResource(wl_resource *resource)
+void OutputDeviceModeV2InterfacePrivate::bindResource(Resource *resource)
 {
-    send_size(resource, m_size.width(), m_size.height());
-    send_refresh(resource, m_refreshRate);
+    send_size(resource->handle, m_size.width(), m_size.height());
+    send_refresh(resource->handle, m_refreshRate);
 
-    if (m_preferred) {
-        send_preferred(resource);
+    if (m_flags & OutputMode::Flag::Preferred) {
+        send_preferred(resource->handle);
+    }
+    if (resource->version() >= KDE_OUTPUT_DEVICE_MODE_V2_FLAGS_CUSTOM) {
+        uint32_t flags = 0;
+        if (m_flags & OutputMode::Flag::Custom) {
+            flags |= KDE_OUTPUT_DEVICE_MODE_V2_FLAGS_CUSTOM;
+        }
+        if (m_flags & OutputMode::Flag::ReducedBlanking) {
+            flags |= KDE_OUTPUT_DEVICE_MODE_V2_FLAGS_REDUCED_BLANKING;
+        }
+        send_flags(resource->handle, flags);
     }
 }
 
